//
//  NSAccessibility+NisusCore.h
//  NisusCore
//
//  Created by Martin Wierschin on 2013.08.28.
//  Copyright (c) 2013 Nisus. All rights reserved.
//
// The Apple way of implementing accessibility is to use a giant set of if/then/else tests against an attribute key.
// That's tedious. Instead, a class may simply use the single define below, and then implement each attribute as a method.
//
// For example, instead of:
// - (id) accessibilityAttributeValue:(NSString*)attribute {
//        if( [attribute isEqualToString:NSAccessibilityRoleAttribute] ) return NSAccessibilityTableRole;
//        ...
//	}
//
// We can do:
// - (id) nisusAccessibilityValue_role {
//       return NSAccessibilityTableRole;
// }
//
//
// Similarly for actions, instead of comparing against NSAccessibilityPressAction, just implement:
// - (void) nisusAccessibilityAction_press ;

#import <Cocoa/Cocoa.h>

// Methods that custom classes might want to override/call.
@interface NSObject (NisusAccessibilityAdditions)
- (id) nisusAccessibilityValueForAttribute:(NSString*)attrName ;
- (id) nisusAccessibilityValue_roleDescription ;
@end


// Checks if super responds.
static inline BOOL NisusSuperRespondsToSelector( Class impClass, SEL selector ) 
{
	return [[impClass superclass] instancesRespondToSelector:selector];
}




// Provides override of Apple method that calls into specific Nisus methods.
#define NTDefineAccessibilityAutoDispatchForClass( implClassName )									\
- (NSArray*) accessibilityAttributeNames															\
{																									\
	NSArray* names = [self _nisusAccessibilityImpl_cachedAttributeNames];							\
	if( nil == names ) {																			\
		NSArray* superNames = nil;																	\
		BOOL isUseSuper = NisusSuperRespondsToSelector( [implClassName class], _cmd );					\
		if( isUseSuper ) superNames = [super accessibilityAttributeNames];							\
		names = [self _nisusAccessibilityImpl_cacheAttributeNamesFromSuper:superNames];				\
	}																								\
	return names;																					\
}																									\
																								\
- (id) accessibilityAttributeValue:(NSString*)attrName											\
{																								\
	id value = [self _nisusAccessibilityImpl_valueForAttribute:attrName];						\
	if( nil == value ) {																		\
		BOOL isUseSuper = NisusSuperRespondsToSelector( [implClassName class], _cmd );				\
		if( isUseSuper ) value = [super accessibilityAttributeValue:attrName];					\
	}																							\
	return value;																				\
}																								\
																								\
- (BOOL) accessibilityIsAttributeSettable:(NSString*)attrName												\
{																											\
	BOOL isSettable = NO;																					\
	if(! [self _nisusAccessibilityImpl_getIsSettable:&isSettable forAttribute:attrName] ) {					\
		BOOL isUseSuper = NisusSuperRespondsToSelector( [implClassName class], _cmd );							\
		if( isUseSuper ) isSettable = [super accessibilityIsAttributeSettable:attrName];					\
	}																										\
	return isSettable;																						\
}																											\
																											\
- (void) accessibilitySetValue:(id)value forAttribute:(NSString*)attrName						\
{																								\
	if(! [self _nisusAccessibilityImpl_setValue:value forAttribute:attrName] ) {				\
		BOOL isUseSuper = NisusSuperRespondsToSelector( [implClassName class], _cmd );				\
		if( isUseSuper ) [super accessibilitySetValue:value forAttribute:attrName];				\
	}																							\
}																									\
																									\
- (NSArray*) accessibilityActionNames																\
{																									\
	NSArray* names = [self _nisusAccessibilityImpl_cachedActionNames];								\
	if( nil == names ) {																			\
		NSArray* superNames = nil;																	\
		BOOL isUseSuper = NisusSuperRespondsToSelector( [implClassName class], _cmd );					\
		if( isUseSuper ) superNames = [super accessibilityActionNames];								\
		names = [self _nisusAccessibilityImpl_cacheActionNamesFromSuper:superNames];				\
	}																								\
	return names;																					\
}																									\
																									\
- (void) accessibilityPerformAction:(NSString*)actionName										\
{																								\
	if(! [self _nisusAccessibilityImpl_performAction:actionName] ) {							\
		BOOL isUseSuper = NisusSuperRespondsToSelector( [implClassName class], _cmd );				\
		if( isUseSuper ) [super accessibilityPerformAction:actionName];							\
	}																							\
}


// These methods are only used by internal implementation (above macro defines)
@interface NSObject (NisusAccessibilityImpl)
- (NSArray*) _nisusAccessibilityImpl_cachedAttributeNames ;
- (NSArray*) _nisusAccessibilityImpl_cacheAttributeNamesFromSuper:(NSArray*)superNames ;

- (id) _nisusAccessibilityImpl_valueForAttribute:(NSString*)attrName ;
- (BOOL) _nisusAccessibilityImpl_getIsSettable:(BOOL*)isSettablePtr forAttribute:(NSString*)attrName ;
- (BOOL) _nisusAccessibilityImpl_setValue:(id)value forAttribute:(NSString*)attrName ;

- (NSArray*) _nisusAccessibilityImpl_cachedActionNames ;
- (NSArray*) _nisusAccessibilityImpl_cacheActionNamesFromSuper:(NSArray*)superNames ;
- (BOOL) _nisusAccessibilityImpl_performAction:(NSString*)actionName ;
@end


