//
//  NisusAssertions.h
//  NisusCore
//
//  Created by Martin Wierschin on 2013.08.28.
//  Copyright (c) 2013 Nisus. All rights reserved.
//
// This is the public header that exposes all NisusCore APIs.
// This framework is intended to be general purpose for reuse in all Nisus products.

#ifndef NISUS_CORE_ASSERTIONS_INCLUDE_H
#define NISUS_CORE_ASSERTIONS_INCLUDE_H 1

#import <NisusCore/NisusDefines.h>


// Whether or not assertions are enabled. If not, turns all asserts into NOOPs.
#if NISUS_DEBUG
	#define NISUS_CORE_ASSERTIONS_ENABLED 1
#else
	#define NISUS_CORE_ASSERTIONS_ENABLED 0
#endif


typedef enum {
	NisusAssertType_Normal,		// failed normal assertion: one that should trigger error reporting
	NisusAssertType_Soft,		// soft assertion, do not trigger error reporting
	NisusAssertType_Exception,	// encountered unexpected runtime exception
	NisusAssertType_Abstract,	// abstract method was called
} NisusAssertType;



#if NISUS_CORE_ASSERTIONS_ENABLED
	
	#ifdef __cplusplus
	extern "C" {
	#endif
		void NisusAssertFailed( NisusAssertType assertType, const char* expression, const char* userFriendlyError, const char* fileName, unsigned lineNumber );
	#ifdef __cplusplus
	}
	#endif

	#ifdef __OBJC__
		void NisusAbstractMethodCalled( id self, SEL selector, const char* fileName, unsigned lineNumber );
	#endif
	
	
	#define _NISUS_ASSERT(assertType, expression, userFriendlyError)										\
	{																										\
		if( ! (expression) ) {																				\
			NisusAssertFailed( assertType, #expression, userFriendlyError, __FILE__, __LINE__ );			\
		}																									\
	}

#else // NISUS_CORE_ASSERTIONS_ENABLED is false

	#define _NISUS_ASSERT(assertType, expression, userFriendlyError)
	#define NisusAbstractMethodCalled(self, selector, fileName, lineNumber)		

#endif



// Use this instead of other ASSERT macros, will signal a serious error, but won't call abort().
// If the expression evaluates to false, will trigger the assertion (log error, trigger feedback reporter, etc).
// The "userFriendlyError" should be some snippet we can show to beta testers, so they have some idea about the area of failure. The message is NOT shown/used in release builds.
// An appropriate string might be something like "zoom in page view" or "can't insert new image", just to give them an idea what feature was implicated.
#define NISUS_ASSERT(expression, userFriendlyError)	\
	_NISUS_ASSERT( NisusAssertType_Normal, expression, userFriendlyError )


// Raw just lets you assert without a friendly user message. Not preferred, as user will be shown failed condition code instead.
#define NISUS_ASSERT_RAW(expression)	\
	_NISUS_ASSERT( NisusAssertType_Normal, expression, "" )

// Soft assertions are less important failures. They trigger some logging, but no user visible effects like the feedback reporter.
// Ideally the developer has a breakpoint set for all breakpoints, soft or not, so they can be corrected.
#define NISUS_ASSERT_SOFT(expression)	\
	_NISUS_ASSERT( NisusAssertType_Soft, expression, "" )

// If the condition is not met, then assert in debug builds, or execute the given safety code in release builds.
#define NISUS_ASSERT_SAFE(cond, userFriendlyError, safetyCode)	\
{																\
	NISUS_ASSERT( cond, userFriendlyError );					\
	if(! (cond) ) { safetyCode; }								\
}

// Used as implementations for abstract methods.
#define NISUS_ASSERT_ABSTRACT \
	NisusAbstractMethodCalled( self, _cmd, __FILE__, __LINE__ )

#define NISUS_ASSERT_ABSTRACT_RETURN( returnValue ) \
{ \
	NISUS_ASSERT_ABSTRACT;	\
	return (returnValue);	\
}






// Cast where we assert that the value is a particular kind of class (or nil).
// Assert returns the casted value of the evaluated expression.
#if NISUS_DEBUG
	#define NISUS_ASSERT_CAST(className, valueExpr)																				\
		({																														\
			id nisusRawValue = (valueExpr);																						\
			NISUS_ASSERT( (nil == nisusRawValue) || ([nisusRawValue isKindOfClass:[className class]]), "expected className" );	\
			className* nisusCastedValue = (className*)(nisusRawValue);															\
			nisusCastedValue;																									\
		})
#else
	#define NISUS_ASSERT_CAST(className, valueExpr)			((className*)(valueExpr))
#endif




#endif // NISUS_CORE_ASSERTIONS_INCLUDE_H

