//
//  NisusMemBlock.h
//  InfoClick
//
//  Created by Martin Wierschin on 2013.03.15.
//  Copyright (c) 2013 __MyCompanyName__. All rights reserved.
//
// This is just a set of macros that sanitizes access to a raw memory buffer.
// It only incurs additional overhead in DEBUG builds, in order to catch errors.
// In a release build you have all the efficiency of unregulated memory access (and the rope to hang yourself).

#import <Foundation/Foundation.h>


#if NISUS_DEBUG
	static inline void _NisusMemBlock_ThrowException( const char* file, NSUInteger fileLine, NSUInteger invalidIndex, NSUInteger capacity )
	{
		[NSException raise:@"NisusMemBlockException" format:@"NisusMemBlock out-of-bounds access (%lu >= %lu), at %s:%ld", (long)invalidIndex, (long)capacity, file, (long)fileLine];
	}

	#define _NisusMemBlock_ENSURE_INDEX_IN_BOUNDS( ivarName, testIndexExpr )	\
		{																		\
			NSUInteger testIndex = testIndexExpr;								\
			if( testIndex >= (ivarName ## _capacity) ) _NisusMemBlock_ThrowException( __FILE__, __LINE__, testIndex, (destBlockName ## _capacity) ); \
		}

#else

	#define _NisusMemBlock_ENSURE_INDEX_IN_BOUNDS( ivarName, testIndexExpr )

#endif




// This needs to wrap the ivar name, eg: scalartype* NisusMemBlock_DECLARE(ivar);
#define NisusMemBlock_DECLARE(ivarName) \
	ivarName;							\
	NSUInteger ivarName ## _capacity;


// Frees the buffer if needed, and set it to a NULL pointer.
#define NisusMemBlock_FREE(ivarName)	\
	if( NULL != ivarName ) {			\
		free(ivarName);					\
		ivarName = NULL;				\
		(ivarName ## _capacity) = 0;	\
	}

// Wholesale replaces the block with another. Frees any existing block and assumes ownership of the given block.
#define NisusMemBlock_SET_NOCOPY(ivarName, newBlock, newCapacity)	\
	NisusMemBlock_FREE(ivarName);									\
	ivarName = newBlock;											\
	(ivarName ## _capacity) = newCapacity;





#define NisusMemBlock_CAPACITY_GET(ivarName) \
	(ivarName ## _capacity)

// Reallocates the buffer to hold the given number of elements. If succeeds, returns YES.
// If the realloc fails, returns NO and leaves the existing buffer intact.
extern BOOL _NisusMemBlock_ReallocImp( void** bufferPtr, NSUInteger* curCapPtr, NSUInteger elSize, NSUInteger newCap );
extern BOOL _NisusMemBlock_ReallocZeroedImp( void** bufferPtr, NSUInteger* curCapPtr, NSUInteger elSize, NSUInteger newCap );

#define NisusMemBlock_CAPACITY_SET(ivarName, newCapacity) \
	(_NisusMemBlock_ReallocImp( (void**)(&ivarName), &(ivarName ## _capacity), sizeof(*ivarName), (newCapacity)))

#define NisusMemBlock_CAPACITY_SET_AND_ZERO(ivarName, newCapacity) \
	(_NisusMemBlock_ReallocZeroedImp( (void**)(&ivarName), &(ivarName ## _capacity), sizeof(*ivarName), (newCapacity)))


// This ensures the memory buffer can hold at least the specified number of elements.
// If the buffer had not yet been allocated, then will allocate it as necessary.
// Returns YES on success, or NO on failure (if fails, then existing buffer/size is kept intact).
static inline BOOL _NisusMemBlock_EnsureCapacityImp( void** bufferPtr, NSUInteger* curCapPtr, NSUInteger elSize, NSUInteger minCap )
{
	NSUInteger curCap = *curCapPtr;
	if( curCap >= minCap ) return YES;
	return _NisusMemBlock_ReallocImp(bufferPtr, curCapPtr, elSize, minCap);
}

#define NisusMemBlock_CAPACITY_ENSURE(ivarName, minCapacity) \
	(_NisusMemBlock_EnsureCapacityImp( (void**)&ivarName, &(ivarName ## _capacity), sizeof(*ivarName), (minCapacity)))




#if NISUS_DEBUG
	// Accesses an element from the buffer. If the index is out of bounds, throws an exception in debug builds.
	static inline void* _NisusMemBlock_GetElementImp( void* buffer, NSUInteger bufferCapacity, NSUInteger elIndex, NSUInteger elSize, const char* file, NSUInteger fileLine )
	{
		if( elIndex >= bufferCapacity ) _NisusMemBlock_ThrowException(file, fileLine, elIndex, bufferCapacity);
		return ((char*)buffer) + (elIndex * elSize);
	}

	#define NisusMemBlock_ELEMENT_GET_PTR(ivarName, elIndex)	\
		({												\
			NSUInteger elIndexVal = (elIndex);				\
			if( elIndexVal >= (ivarName ## _capacity) ) _NisusMemBlock_ThrowException( __FILE__, __LINE__, elIndexVal, (ivarName ## _capacity) ); \
			&(ivarName[elIndexVal]);						\
		})
#else
	#define NisusMemBlock_ELEMENT_GET_PTR(ivarName, elIndex) \
		(&(ivarName[elIndex]))
#endif


#define NisusMemBlock_ELEMENT_GET(ivarName, elIndex) \
	(*NisusMemBlock_ELEMENT_GET_PTR(ivarName, elIndex))


#define NisusMemBlock_ELEMENT_SET(ivarName, elIndex, value) \
	(*NisusMemBlock_ELEMENT_GET_PTR(ivarName, elIndex)) = (value)

// Sets the memory of the block to the given block, assuming the same data type. Leaves the capacity unchanged.
#if NISUS_DEBUG
	#define NisusMemBlock_ELEMENTS_SET(destBlockName, destBaseIndex, sourceBlock, sourceCountExpr)	\
		{																						\
			NSUInteger sourceCount = sourceCountExpr;											\
			if( sourceCount > 0 ) {																\
				NSUInteger lastSetIndex = destBaseIndex + sourceCount - 1;						\
				if( lastSetIndex >= (destBlockName ## _capacity) ) _NisusMemBlock_ThrowException( __FILE__, __LINE__, lastSetIndex, (destBlockName ## _capacity) ); \
				memmove( destBlockName + destBaseIndex, sourceBlock, (sourceCount * sizeof(*destBlockName)) ); \
			}																					\
		}

#else

	#define NisusMemBlock_ELEMENTS_SET(destBlockName, destBaseIndex, sourceBlock, sourceCountExpr)	\
		{																						\
			NSUInteger sourceCount = sourceCountExpr;											\
			if( sourceCount > 0 ) {																\
				memmove( destBlockName + destBaseIndex, sourceBlock, (sourceCount * sizeof(*destBlockName)) ); \
			}																					\
		}

#endif


