//
//  NisusMutationState.h
//  WriterMerged
//
//  Created by Martin on 2011.03.04.
//  Copyright 2011 Nisus. All rights reserved.
//
// To help conveniently manage mutation when efficiency is important

#import <Foundation/Foundation.h>
#import <NisusCore/NisusDefines.h>
#import <NisusCore/NisusAssertions.h>

typedef id (*NisusCreateMutableObjectFunc)();

typedef struct {
	id original;
	id mutated;
	NisusCreateMutableObjectFunc allocFunc;
} NisusMutationState;



// Helper functions for mutable state
NSMutableDictionary* NisusCreateMutableDictionary();
NSMutableArray* NisusCreateMutableArray();
NSMutableSet* NisusCreateMutableSet();
id NisusMutationState_AllocAndSetMutableObject( NisusMutationState* state );
void NisusMutationState_SetMutableObject( NisusMutationState* state, id mobj );

// Returns a new state, which must be disposed. The factory is used in case the original object is nil on creation, and must return
// a new mutable object instance that has a retain count of 1, eg: behaves like [[Object alloc] init].
static NISUS_INLINE NisusMutationState NisusMutationState_Create( id originalObj, NisusCreateMutableObjectFunc allocFunc )
{
	NISUS_ASSERT( ((nil != originalObj) || (NULL != allocFunc)), "mutation state" );
	NisusMutationState state = {[originalObj retain], nil, allocFunc};
	return state;
}

// For clients that already have a mutable object they'd like to modify.
static NISUS_INLINE NisusMutationState NisusMutationState_CreateWithMutableObject( id mobj )
{
	NISUS_ASSERT( nil != mobj, "mutation state" );
	NisusMutationState state = {[mobj retain], [mobj retain], NULL};
	return state;
}

static NISUS_INLINE void NisusMutationState_Dispose( NisusMutationState* state )
{
	[state->original release];
	[state->mutated release];
#if NISUS_DEBUG
	state->original = (id)0xDEADBEEF;
	state->mutated = (id)0xDEADBEEF;
#else
	state->mutated = nil;
#endif
}

// Returns the current object in a way that ensures it will still be valid after disposing of the state.
static NISUS_INLINE id NisusMutationState_DisposeAndReturnCurrentObject( NisusMutationState* state )
{
	id mobj = state->mutated;
	id obj = (nil != mobj) ? [[mobj retain] autorelease] : state->original; // mutated object is released in disposal
	NisusMutationState_Dispose(state);
	return obj;
}

// Returns the current object value, assuming the mutated version is more recent, if it exists.
static NISUS_INLINE id NisusMutationState_CurrentObject( NisusMutationState* state )
{
	id mobj = state->mutated;
	return (nil != mobj) ? mobj : state->original;
}

// Returns the mutable object, creating one as necessary.
static NISUS_INLINE id NisusMutationState_MutableObject( NisusMutationState* state )
{
	id mobj = state->mutated;
	if( nil == mobj ) mobj = NisusMutationState_AllocAndSetMutableObject(state);
	return mobj;
}

// Returns YES if a mutable object was ever created
static NISUS_INLINE BOOL NisusMutationState_WasMutated( NisusMutationState* state )
{
	return (nil != state->mutated);
}

