//
//  NisusObserverSet.h
//  WriterMerged
//
//  Created by Martin on 2006.09.21.
//  Copyright 2006 Nisus Software, Inc. All rights reserved.

// Represents a set of observers that can be notified en-mass of object changes. Observers are not retained by the set!

#import <Cocoa/Cocoa.h>
@class NisusWeakObjectSet, NisusWeakKeyDictionary;



@interface NisusObserverSet : NSObject

- (void) invalidate ;

- (NSUInteger) count ;
- (void) visitObserversUsing:(void(^)(id obs))visitor ;
- (void) visitObserversRespondingToSelector:(SEL)selector using:(void(^)(id obs))visitor ;

// changing observers. NOTE: observers are not retained by the set!
- (void) addObserver:(id)obj ;
- (void) removeObserver:(id)obj ;
- (BOOL) containsObserver:(id)obj ;

// notifying observers
- (void) makeObserversPerformSelector:(SEL)sel ;
- (void) makeObserversPerformSelector:(SEL)sel withObject:(id)obj ;
- (void) makeObserversPerformSelector:(SEL)sel withObject:(id)obj withObject:(id)obj ;
- (void) makeObserversPerformSelector:(SEL)sel withObject:(id)obj withObject:(id)obj withObject:(id)obj ;

@end




//####### Keyed Observations ########
@interface NisusKeyedObserverSet : NisusObserverSet {
	NisusWeakKeyDictionary* _observerMap;
}

- (void) addObserver:(id)obj key:(NSString*)key notifyUsingSelector:(SEL)action ;

- (NSSet*) observersForKey:(NSString*)key ;

- (void) notifyObserversForKey:(NSString*)key withObject:(id)obj ;
- (void) notifyObserversForKeys:(NSSet*)keys withObject:(id)obj ;

@end




// These macros are provided so each object holding a NisusObserverSet doesn't rewrite the same boilerplate methods.
// In your -addObsever:(id<NTWhatever>)obs method just insert this macro.
// It assumes your object defines an -observers property, but that's it. The set of observers is created lazily as needed.
#define NisusObserverSet_ADD_OBSERVER_IMPL( addObsExpr )		\
	{															\
		id addObsValue = (addObsExpr);							\
		NISUS_ASSERT( nil != addObsValue, "observers" );		\
		NISUS_ASSERT( self != addObsValue, "observers" );		\
		if( nil != addObsValue ) {								\
			NisusObserverSet* observers = [self observers];		\
			if( nil == observers ) {							\
				observers = [[NisusObserverSet alloc] init];	\
				[self setObservers:observers];					\
				[observers release];							\
				observers = [self observers];					\
			}													\
			[observers addObserver:(addObsValue)];				\
		}														\
	}


#define NisusObserverSet_REMOVE_OBSERVER_IMPL( remObs )		\
	{														\
		NisusObserverSet* observers = [self observers];		\
		[observers removeObserver:(remObs)];				\
		if( 0 == [observers count] ) {						\
			[self setObservers:nil];						\
		}													\
	}

