//
//  NisusRangeIterator.h
//  NisusCore
//
//  Created by Martin on 2017.12.11.
//  Copyright © 2017 Nisus. All rights reserved.
//
//	This data structure provides an easy way to iterate a range via index.

#import <Foundation/Foundation.h>
#import <NisusCore/NisusRangeUtil.h>


typedef struct {
	NSUInteger nextIndex;
	NSUInteger boundIndex;
	BOOL isForwards;
} NisusRangeIterator;




// Constructs a new iterator that will visit the given range.
// ie: calling NisusRangeIterator_NextRange() will return the given range.
static NISUS_INLINE NisusRangeIterator NisusRangeIterator_Make( NSRange range, BOOL isForwards )
{
	NisusRangeIterator iter;
	iter.isForwards = isForwards;
	iter.nextIndex = isForwards ? range.location : NSMaxRange(range);
	iter.boundIndex = isForwards ? NSMaxRange(range) : range.location;
	return iter;
}

// Returns YES if the given iterator is exhausted and there is no remaining range.
static NISUS_INLINE BOOL NisusRangeIterator_IsEmpty( NisusRangeIterator const * const iter )
{
	if( iter->isForwards ) {
		return (iter->nextIndex >= iter->boundIndex);
	}
	else {
		return (iter->nextIndex <= iter->boundIndex);
	}
}

// Returns the range left in the iterator that hasn't been skipped yet.
// Returns a zero length range if already exhausted.
static NISUS_INLINE NSRange NisusRangeIterator_NextRange( NisusRangeIterator const * const iter )
{
	NSRange remRange;
	
	if( iter->isForwards ) {
		if( iter->nextIndex >= iter->boundIndex ) {
			remRange = NSMakeRange( iter->boundIndex, 0 );
		}
		else {
			remRange = NisusRange_AtWithBound( iter->nextIndex, iter->boundIndex );
		}
	}
	else {
		if( iter->nextIndex <= iter->boundIndex ) {
			remRange = NSMakeRange( iter->boundIndex, 0 );
		}
		else {
			remRange = NisusRange_AtWithBound( iter->boundIndex, iter->nextIndex );
		}
	}
	
	return remRange;
}

static NISUS_INLINE BOOL NisusRangeIterator_GetNextRange( NisusRangeIterator const * const iter, NSRange* nextRangePtr )
{
	NSRange nextRange = NisusRangeIterator_NextRange(iter);
	*nextRangePtr = nextRange;
	return (nextRange.length > 0);
}

// Returns the first index in the next range, or NSNotFound if the iterator is empty.
static NISUS_INLINE NSUInteger NisusRangeIterator_NextIndex( NisusRangeIterator const * const iter )
{
	if( iter->isForwards ) {
		if( iter->nextIndex >= iter->boundIndex ) {
			return NSNotFound;
		}
		else {
			return iter->nextIndex;
		}
	}
	else {
		if( iter->nextIndex <= iter->boundIndex ) {
			return NSNotFound;
		}
		else {
			return iter->nextIndex - 1;
		}
	}
}

// Advances the iterator so it skips the given range, ie: moves to NSMaxRange(skipRange)
static NISUS_INLINE void NisusRangeIterator_SkipRange( NisusRangeIterator* iter, NSRange skipRange )
{
	if( iter->isForwards ) {
		iter->nextIndex = NSMaxRange(skipRange);
	}
	else {
		iter->nextIndex = skipRange.location;
	}
}

// Sets the iterator so the next range it returns includes the given location.
// If the given location is outside the original range then this sets the iterator to empty.
static NISUS_INLINE void NisusRangeIterator_SetNextIndex( NisusRangeIterator* iter, NSUInteger setIndex )
{
	if( iter->isForwards ) {
		iter->nextIndex = setIndex;
	}
	else {
		iter->nextIndex = setIndex + 1;
	}
}


