//
//  NisusSecurityScopedURL.h
//  NisusCore
//
//  Created by Martin Wierschin on 2012.06.01.
//  Copyright (c) 2012 Nisus. All rights reserved.
//
//
// Class that helps isolate the new security-scoped URL access required for sandboxing under 10.7.3.
// It transparently handles capturing bookmark data and whatever else might be necessary to keep a URL viable between app launches.
//
// NOTE: instances of this class ARE thread-safe.

#import <Foundation/Foundation.h>
@class NisusSecureURLSerialization;


typedef enum {
	NisusSecureAccessFlag_None			= 0,
	NisusSecureAccessFlag_AllowGUI		= 1 << 0,
	NisusSecureAccessFlag_AllowMount	= 1 << 1,
} NisusSecureAccessFlags;

#define NisusSecureAccessFlags_USER_ACTION (NisusSecureAccessFlag_AllowGUI | NisusSecureAccessFlag_AllowMount)

extern NSString* NisusSecureURL_AliasFileExtension;



@interface NisusSecureURL : NSObject <NSCopying> {
	NSError* _initError;
	NSURL* _resolvedURL;
	NSMutableArray* _openTokens;
	NSURL* _staleURL;
}

+ (BOOL) isSecurityScopingViableAndGetError:(NSError**)errorPtr ;

// creating app-scoped URLs
- (instancetype) initCapturingBookmarkForURL:(NSURL*)url readOnly:(BOOL)isReadOnly error:(NSError**)errorPtr ;
- (instancetype) initUsingParentFolderSecureURL:(NisusSecureURL*)parentSecureURL targetURL:(NSURL*)targetURL error:(NSError**)errorPtr ;
- (instancetype) initWithSerializedData:(NSData*)bmarkData ;

// creating document-scoped URLs, where the bookmark will be stored inside a document with the given URL
- (instancetype) initDocumentScopedCapturingBookmarkForURL:(NSURL*)url readOnly:(BOOL)isReadOnly inDocumentWithURL:(NSURL*)docURL error:(NSError**)error ;
- (instancetype) initDocumentScopedWithSerializedData:(NSData*)bmarkData inDocumentWithURL:(NSURL*)docURL ;

// properties
- (NSData*) serializationStorableGlobally ;
- (NSData*) serializationStorableInDocumentWithURL:(NSURL*)docURL ;

- (NSString*) unresolvedFileName ;
- (NSString*) unresolvedFilePath ;
- (NSURL*) lastResolvedURL ;

// convenience methods that do not resolve the bookmark, will use last known info
- (NSURL*) fileURL ;
- (NSString*) filePath ;
- (NSString*) fileName ;

// resolving the bookmark data to a URL, and security-scope access thereof
- (BOOL) isAccessibleWithFlags:(NisusSecureAccessFlags)flags ;
- (BOOL) accessWithFlags:(NisusSecureAccessFlags)flags getError:(NSError**)errorPtr using:(void(^)(NSURL* url))visitor ;
- (BOOL) accessWithFlags:(NisusSecureAccessFlags)flags getError:(NSError**)errorPtr asyncUsing:(void(^)(NSURL* url, void(^completion)() ))visitor ;
- (BOOL) accessWithFlags:(NisusSecureAccessFlags)flags errorPresenter:(id)delegate using:(void(^)(NSURL* url))visitor ;

- (NSURL*) startAccessingWithFlags:(NisusSecureAccessFlags)flags error:(NSError**)errorPtr ;
- (void) stopAccessing ;

- (NSURL*) resolvedURL ;
- (NSURL*) resolvedURLWithFlags:(NisusSecureAccessFlags)flags error:(NSError**)errorPtr ;
- (void) clearResolvedURL ;

// Finder aliases
- (BOOL) writeFinderAliasToURL:(NSURL*)writeURL error:(NSError**)errorPtr ;
+ (NisusSecureURL*) readFinderAliasFromURL:(NSURL*)readURL error:(NSError**)errorPtr ;

// operations
- (void) setCorrectedSecureURL:(NisusSecureURL*)newSecureURL ;
- (NisusSecureURL*) secureURLForSubfolderWithName:(NSString*)subfolderName create:(BOOL)isCreate error:(NSError**)errorPtr ;

@end





/*** UserDefaults ***/
@interface NSUserDefaults (NisusSecureURL)

- (NisusSecureURL*) nisusSecureURLForKey:(NSString*)key ;
- (void) setNisusSecureURL:(NisusSecureURL*)url forKey:(NSString*)key ;

@end



/*** Debugging Tools ***/
#if NISUS_DEBUG
	extern BOOL NisusSecureURL_IsLogSandboxActivity;

	#define NisusSecureURL_LogSandboxActivity( str, args... )		\
		if( NisusSecureURL_IsLogSandboxActivity ) {					\
			NSLog(str , ## args);									\
		}

#else
	// logging is off/NOOP
	#define NisusSecureURL_LogSandboxActivity( str, args... )

#endif // NISUS_DEBUG
