//
//  NisusWeakReference.h
//  Toolbox
//
//  Created by Martin Wierschin on Mon Sep 27 2004.
//  Copyright (c) 2004 Nisus Software, Inc. All rights reserved.

// Holds a reference to an object without retaining the object. This is useful for putting
// objects in collections without retaining them. This class also has a few benefits over
// NSValues nonretainedObjectValue:
//
// 1. More efficient. Not only are instances pooled, but retainCount is stored inline.
// 2. Sanity check to prevent accessing deallocated objects.

#import <Foundation/Foundation.h>




@interface NisusWeakReference : NSObject <NSCopying> {
	NSObject* NISUS_WEAK_IVAR_DECLARE(_weakObj);
}

- (id) initWithObject:(NSObject*)obj ;

- (id) object ;
- (void) setObject:(NSObject*)obj ;

@end



// Additions that make managings sets of weak references easier and more efficient.
// (Main efficiency is that weak ref object can be returned to shared pool right after operation).
@interface NSArray (NisusWeakReferenceAdditions)
- (NSUInteger) indexOfWeakReferenceObject:(id)obj ;
- (BOOL) containsWeakReferenceObject:(id)obj ;
@end

@interface NSMutableArray (NisusWeakReferenceAdditions)
- (void) addWeakReferenceToObject:(id)obj ;
- (void) removeWeakReferenceToObject:(id)obj ;
@end

@interface NSSet (NisusWeakReferenceAdditions)
- (BOOL) containsWeakReferenceObject:(id)obj ;
@end

@interface NSMutableSet (NisusWeakReferenceAdditions)
- (void) addWeakReferenceObject:(id)obj ;
- (void) removeWeakReferenceObject:(id)obj ;
@end

@interface NSMutableDictionary (NisusWeakReferenceAdditions)
- (void) removeObjectForWeakKey:(id)key ;
- (void) setObject:(id)obj forWeakKey:(id)key ;
- (id) objectForWeakKey:(id)key ;
@end

